(define-module (alterator workflow form)
  :use-module (srfi srfi-1)
  :use-module (srfi srfi-13)

  :use-module (alterator ahttpd)
  :use-module (alterator ahttpd html)

  :use-module (alterator plist)
  :use-module (alterator algo)
  :use-module (alterator str)
  :use-module (alterator woo)
  :use-module (alterator http template)
  :use-module (alterator http html)
 :export (;;high level
          workflow
          fill-form))

;;; main form processing

(define (not-name? name)
  (lambda(x) (not (eq? (car x) name))))

(define (redirect-args name-attribute url-args)
  (let ((name-value (cond-assq 'name url-args)))
    (if (and name-attribute name-value)
      (acons name-attribute name-value
             (filter (not-name? name-attribute) url-args))
      url-args)))

(define (redirect-form url url-args template-args)
  (let* ((async (cond-assq 'async url-args))
	 (redirect-url (or (cond-assq 'redirect template-args)
			   (cond-assq 'redirect-url template-args)
			   url))
	 (name-attribute (cond-assq 'name-attribute template-args))
	 (url-args (redirect-args name-attribute url-args)))
    (if (or (string-null? redirect-url) async name-attribute)
      (workflow-form url "read" url-args template-args)
      (html:redirect redirect-url))))

(define (filter-action url-args)
  (plist-remove (lambda (x y) (eq? x 'action)) url-args))

(define (workflow-form url action url-args template-args)
  (let ((scm (cond-assq 'scm template-args))
	(backend-url (cond-assq 'backend-url template-args url))
	(url-redirect (cond-assq 'url-redirect template-args #f))
	(backend-args (backend-args url-args)))
    (cond
      (url-redirect
	(html:redirect url-redirect))
      ((string=? action "read")
	 (template scm (apply fill-form backend-url (filter-action backend-args))))
      (else
        (apply woo action backend-url backend-args)
	(redirect-form url url-args template-args)))))

(define (ajax-form scm)
  (template
    scm
    (replace-tag: "body" (lambda(options content)
			   `(body ,@options
				  (script ,(format #f "$(document).ready(function(){ form_callback('init'); });"))
				  ,@content)))))

(define (workflow url msg template-args)
  (let ((action (message-action msg))
        (url-args (message-woo-args msg)))
    (ajax-form
     (catch/message
      (lambda() (workflow-form url action url-args template-args))
      (lambda() (workflow-form url "read" url-args template-args))))))

;;; form fill

(define (fill-form obj-url . obj-args)
  (process-form (delay (apply woo-read-first obj-url obj-args))
		obj-url obj-args))

(define (process-form obj obj-url obj-args)
    (list
     ;;anchors
     (process-a obj obj-url obj-args)

     ;;spans
     (process-span obj)

     ;;images
     (process-img obj)

     ;; selects
     (process-select obj obj-url obj-args)

     ;; inputs
     (process-hidden obj)
     (process-checkbox obj)
     (process-edit obj)

     ;; textareas
     (process-textarea obj)

     ;; tables
     (process-listbox2 obj-url obj-args) ))

;;; helpers

(define (fill-content content obj obj-url obj-args)
  (template
    content
    (process-form obj obj-url obj-args)))

(define (get-value options obj . default)
  (let* ((name (cond-assq 'name options ""))
         (prev-value (cond-assq 'value options ""))
         (default (if (pair? default) (car default) prev-value)))
    (woo-get-option (force obj) (string->symbol name) default)))

(define (tag-options options)
 (filter (lambda(x) (not (memq (car x) '(value href checked))))
         options))

;;; selects

(define (select-url obj-url options)
  (let ((optionlist-attribute (cond-assq 'optionlist options))
	(enumref-attribute (cond-assq 'enumref options)))
    (or enumref-attribute
	(and (string? optionlist-attribute)
	     (string-append obj-url "/" optionlist-attribute)))))

(define (process-select obj obj-url obj-args)
  (replace-tag: "select"
		(lambda (options content)
		  (let* ((variant-url (select-url obj-url options)))
		    (if (string? variant-url)
		      `(select ,@options
			       ,(html:select-options (get-value options obj)
						     (apply woo-list variant-url obj-args)))
		      `(select ,@options ,@content))))))

;;; anchors

(define (process-a obj obj-url obj-args)
  (list
    (replace-tag: "a"
		  (@ 'class "alterator-ref2")
		  (lambda (options content)
		    (let* ((name (string->symbol (cond-assq 'name options "name")))
			   (encode (cond-assq 'encode options "yes"))
			   (ref-obj ((if (string=? encode "yes") encode-url-component values)
				     (woo-get-option (force obj) name)))
			   (ref-url (cond-assq 'href options "")))
		      `(a ,(@ 'href (string-append ref-url ref-obj))
			  ,@(tag-options options)
			  ,(fill-content content obj obj-url obj-args)))))
    (replace-tag: "a"
		  (@ 'class "alterator-ref")
		  (lambda(options content)
		    (let ((ref-obj (car (force obj)))
			  (ref-url (cond-assq 'href options)))
		      `(a ,(@ 'href (string-append ref-obj ref-url))
			  ,@(tag-options options)
			  ,(fill-content content obj obj-url obj-args)))))))

;;; spans

(define (process-span obj)
  (replace-tag: "span"
                (@ 'class "alterator-label")
                (lambda(options content)
		  `(span ,@options ,(->string (get-value options obj))))))


;;; images

(define *img-ext-list* '("png" "jpg" "gif"))
(define *img-data-dir* "/design/images/")
(define *alt-data-dir* "/usr/share/alterator/")
(define *img-null-file* (string-append *img-data-dir* "null.gif"))

(define (img-path src value)
  (let ((prefix (string-append *img-data-dir* (string-trim src #\/) value)))
    (or (any (lambda(ext)
	       (let ((path (string-append prefix "." ext)))
		 (and (read-access? (string-append *alt-data-dir* path))
		      path)))
	     *img-ext-list*)
	*img-null-file*)))

(define (process-img obj)
  (replace-tag: "img"
                (@ 'class "alterator-img")
                (lambda(options content)
		  `(img ,(@ 'src (img-path (cond-assq 'src options "")
		                           (->string (get-value options obj))))))))

;;; inputs

(define (form-error . msg)
  (throw 'form-error (apply format #f msg)))

(define (process-hidden obj)
  (replace-tag: "input"
		(@ 'type "hidden")
		(lambda (options content)
		  (let ((name (cond-assq 'name options "unknown"))
			(value (get-value options obj)))
		    (cond
		      ((or (string? value) (number? value))
		       `(input ,@(tag-options options)
			       (value . ,(->string value))))
		      (else
			(form error "hidden input (~S): expected string or number, got ~S" name value)))))))

(define (process-checkbox obj)
  (replace-tag: "input"
		(@ 'type "checkbox")
		(lambda (options content)
		  (let* ((name (cond-assq 'name options "unknown"))
			 (prev-checked (string=? "yes" (cond-assq 'checked options  "no")))
			 (value (get-value options obj prev-checked)))
		    (cond
		      ((boolean? value)
		       `(input ,@(tag-options options)
			       ,(if value (@ 'checked "yes") "")
			       (value . "#t")))
		      ((string? value)
		       `(input ,@(tag-options options)
			       (value . ,value)))
		      (else
			(form-error "checkbox (~S): expected string or boolean, got ~S" name value)))))))

(define (process-edit obj)
  (replace-tag: "input"
		(@ 'type "text")
		(lambda(options content)
		  (let ((name (cond-assq 'name options "unknown"))
			(value (get-value options obj)))
		    (cond
		      ((or (string? value) (number? value))
		       `(input ,@(tag-options options)
			       (value . ,(->string value))))
		      (else
			(form-error "text input (~S): expected string or number, got ~S" name value)))))))

;;; textareas

(define (process-textarea obj)
  (replace-tag: "textarea"
                (lambda(options content)
		  (let ((name (cond-assq 'name options "unknown"))
			(value (get-value options obj)))
		    (cond
		      ((string? value)
		       `(textarea ,@options ,value))
		      (else
			(form-error "text area (~S): expected string, got ~S" name value)))))))

;;; tables

(define (process-listbox2-content scm obj-url obj-args variant-list)
  (template
    scm
    (replace-tag: "tbody"
		  (lambda(options content)
		    `(tbody ,@options
			    ,@(map (lambda(obj)
				     (fill-content content
						   (delay obj)
						   obj-url
						   obj-args))
				   variant-list))))))

(define (process-listbox2 obj-url obj-args)
  (replace-tag: "table"
		(@ 'class "alterator-listbox")
		(lambda(options content)
		  (let* ((variant-url (cond-assq 'enumref options ""))
			 (variant-list (apply woo-list variant-url obj-args)))
		    `(table ,@options
			    ,@(process-listbox2-content content
							obj-url
							obj-args
							variant-list))))))
